import numpy as np

import pyopencl as cl  # noqa
from pyopencl.tools import (
    pytest_generate_tests_for_pyopencl as pytest_generate_tests,  # noqa
)

import loopy as lp


def test_laplacian_stiffness(ctx_factory):
    dtype = np.float32
    ctx = ctx_factory()
    order = "C"

    dim = 2  # (baked into code)

    Nq = 40  # num. quadrature points (baked into code)  # noqa
    Nb = 20  # num. basis functions (baked into code)  # noqa
    Nc = 100  # num. cells (run-time symbolic)  # noqa

    from pymbolic import var
    Nc_sym = var("Nc")  # noqa

    knl = lp.make_kernel(ctx.devices[0],
            "[Nc] -> {[K,i,j,q, dx_axis, ax_b]: 0<=K<Nc and 0<=i,j<%(Nb)d and 0<=q<%(Nq)d "  # noqa
            "and 0<= dx_axis, ax_b < %(dim)d}"
            % {"Nb": Nb, "Nq": Nq, "dim": dim},
            [
                "dPsi(ij, dxi) := sum_float32(@ax_b,"
                    "  jacInv[ax_b,dxi,K,q] * DPsi[ax_b,ij,q])",
                "A[K, i, j] = sum_float32(q, w[q] * jacDet[K,q] * ("
                    "sum_float32(dx_axis, dPsi$one(i,dx_axis)*dPsi$two(j,dx_axis))))"
                ],
            [
            lp.GlobalArg("jacInv", dtype, shape=(dim, dim, Nc_sym, Nq), order=order),
            lp.ConstantArg("DPsi", dtype, shape=(dim, Nb, Nq), order=order),
            lp.GlobalArg("jacDet", dtype, shape=(Nc_sym, Nq), order=order),
            lp.ConstantArg("w", dtype, shape=(Nq,), order=order),
            lp.GlobalArg("A", dtype, shape=(Nc_sym, Nb, Nb), order=order),
            lp.ValueArg("Nc",  np.int32, approximately=1000),
            ],
            name="lapquad", assumptions="Nc>=1")

    knl = lp.tag_inames(knl, {"ax_b": "unr"})
    seq_knl = knl

    def variant_fig31(knl):
        # This (mostly) reproduces Figure 3.1.

        knl = lp.tag_inames(knl, {"dx_axis": "unr"})
        return knl, ["K", "i", "j", "q", "ax_b_insn"]

    def variant_pg4(knl):
        # This (mostly) reproduces the unlabeled code snippet on pg. 4.

        knl = lp.tag_inames(knl, {"dx_axis": "unr"})
        Ncloc = 16  # noqa
        knl = lp.split_iname(knl, "K", Ncloc,
                outer_iname="Ko", inner_iname="Kloc")
        return knl, ["Ko", "Kloc", "i", "j", "q", "ax_b_insn"]

    def variant_fig32(knl):
        # This (mostly) reproduces Figure 3.2.

        Ncloc = 16  # noqa
        knl = lp.split_iname(knl, "K", Ncloc,
                outer_iname="Ko", inner_iname="Kloc")
        knl = lp.precompute(knl, "dPsi", np.float32, ["i", "q", "dx_axis"],
                default_tag=None)
        knl = lp.tag_inames(knl, {"dx_axis": "unr", "dxi": "unr"})
        return knl, ["Ko", "Kloc", "dPsi_q", "ij", "i", "j", "q", "ax_b_insn"]

    def variant_fig33(knl):
        # This is meant to (mostly) reproduce Figure 3.3.

        Ncloc = 16  # noqa
        knl = lp.split_iname(knl, "K", Ncloc,
                outer_iname="Ko", inner_iname="Kloc")
        knl = lp.precompute(knl, "dPsi$one", np.float32, ["dx_axis"], default_tag=None)
        knl = lp.tag_inames(knl, {"j": "ilp.seq"})

        return knl, ["Ko", "Kloc"]

    def variant_simple_gpu(knl):
        # This is a simple GPU-ish variant.

        # It's not the same thing as Matt's code, but I'll need some more time
        # to reverse-engineer what is going on there. Some discussion might
        # help, too. :)

        knl = lp.tag_inames(knl, {"dx_axis": "unr"})
        Ncloc = 16  # noqa
        knl = lp.split_iname(knl, "K", Ncloc,
                outer_iname="Ko", inner_iname="Kloc",
                outer_tag="g.0")
        knl = lp.tag_inames(knl, {"i": "l.1", "j": "l.0"})
        return knl, ["K", "i", "j", "q", "ax_b_insn"]

    def variant_simple_gpu_prefetch(knl):
        # This adds prefetching to the GPU variant above.

        # In this variant (on my machine), loopy makes a silly choice
        # for the upper bound of Kloc (it uses Nc). I'll investigate and
        # fix that. (FIXME)

        knl = lp.tag_inames(knl, {"dx_axis": "unr"})
        Ncloc = 16  # noqa
        knl = lp.split_iname(knl, "K", Ncloc,
                outer_iname="Ko", inner_iname="Kloc",
                outer_tag="g.0")
        knl = lp.tag_inames(knl, {"i": "l.1", "j": "l.0"})
        knl = lp.add_prefetch(knl, "w", ["q"], default_tag="l.auto")
        knl = lp.add_prefetch(knl, "DPsi", [0, 1, 2], default_tag="l.auto")
        knl = lp.add_prefetch(knl, "jacInv", [0, 1, 3], default_tag="l.auto")
        knl = lp.add_prefetch(knl, "jacDet", [1], default_tag="l.auto")
        return knl, ["K", "i", "j", "q", "ax_b_insn"]

    # Plug in variant name here
    #                        |
    #                        v
    for variant in [variant_fig33]:
        var_knl, loop_prio = variant(knl)
        kernel_gen = lp.generate_loop_schedules(var_knl,
                loop_priority=loop_prio)
        kernel_gen = lp.check_kernels(kernel_gen, {"Nc": Nc})

        # print lp.preprocess_kernel(var_knl)

        lp.auto_test_vs_ref(seq_knl, ctx, kernel_gen,
                op_count=0, op_label="GFlops",
                parameters={"Nc": Nc}, print_ref_code=True)


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from py.test.cmdline import main
        main([__file__])



__copyright__ = "Copyright (C) 2015 James Stevens"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging

import numpy as np

from pymbolic.primitives import Variable
from pyopencl.tools import (  # noqa: F401
    pytest_generate_tests_for_pyopencl as pytest_generate_tests,
)
from pytools import div_ceil

import loopy as lp
from loopy.statistics import CountGranularity as CG
from loopy.types import to_loopy_type
from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2  # noqa: F401


logger = logging.getLogger(__name__)

SGS = 32  # Subgroup size


def test_op_counter_basic():

    knl = lp.make_kernel(
            "[n,m,ell] -> {[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
                c[i, j, k] = a[i,j,k]*b[i,j,k]/3.0+a[i,j,k]
                e[i, k+1] = -g[i,k]*h[i,k+1]
                """
            ],
            name="basic", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(knl,
                                  {"a": np.float32, "b": np.float32,
                                       "g": np.float64, "h": np.float64})
    op_map = lp.get_op_map(knl, subgroup_size=SGS, count_redundant_work=True,
                           count_within_subscripts=True)
    n_workgroups = 1
    group_size = 1
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}
    f32add = op_map[lp.Op(np.float32, "add", CG.SUBGROUP, "basic")].eval_with_dict(
            params)
    f32mul = op_map[lp.Op(np.float32, "mul", CG.SUBGROUP, "basic")].eval_with_dict(
            params)
    f32div = op_map[lp.Op(np.float32, "div", CG.SUBGROUP, "basic")].eval_with_dict(
            params)
    f64mul = op_map[lp.Op(np.dtype(np.float64), "mul", CG.SUBGROUP, "basic")
                    ].eval_with_dict(params)
    i32add = op_map[lp.Op(np.dtype(np.int32), "add", CG.SUBGROUP, "basic")
                    ].eval_with_dict(params)
    # (count-per-sub-group)*n_subgroups
    assert f32add == f32mul == f32div == n*m*ell*n_subgroups
    assert f64mul == n*m*n_subgroups
    assert i32add == n*m*2*n_subgroups


def test_op_counter_reduction():

    knl = lp.make_kernel(
            "{[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                "c[i, j] = sum(k, a[i, k]*b[k, j])"
            ],
            name="matmul_serial", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(knl, {"a": np.float32, "b": np.float32})
    op_map = lp.get_op_map(knl, subgroup_size=SGS, count_redundant_work=True)
    n_workgroups = 1
    group_size = 1
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}
    f32add = op_map[lp.Op(np.float32, "add", CG.SUBGROUP,
        "matmul_serial")].eval_with_dict(params)
    f32mul = op_map[lp.Op(np.dtype(np.float32), "mul", CG.SUBGROUP,
        "matmul_serial")].eval_with_dict(params)
    # (count-per-sub-group)*n_subgroups
    assert f32add == f32mul == n*m*ell*n_subgroups

    op_map_dtype = op_map.group_by("dtype")
    f32 = op_map_dtype[lp.Op(dtype=np.float32)].eval_with_dict(params)
    assert f32 == f32add + f32mul


def test_op_counter_logic():

    knl = lp.make_kernel(
            "{[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
                e[i,k] = \
                        (g[i, k] * 2) \
                        if (not(k<ell-2) and k>6 or k/2==ell) else \
                        (g[i, k] + h[i, k] / 2)
                """
            ],
            name="logic", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(knl, {"g": np.float32, "h": np.float64})
    op_map = lp.get_op_map(knl, subgroup_size=SGS, count_redundant_work=True)
    n_workgroups = 1
    group_size = 1
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}
    f32mul = op_map[lp.Op(np.float32, "mul", CG.SUBGROUP, "logic")].eval_with_dict(
            params)
    f64add = op_map[lp.Op(np.float64, "add", CG.SUBGROUP, "logic")].eval_with_dict(
            params)
    f64div = op_map[lp.Op(np.dtype(np.float64), "div", CG.SUBGROUP, "logic")
                    ].eval_with_dict(params)
    i32add = op_map[lp.Op(np.dtype(np.int32), "add", CG.SUBGROUP, "logic")
                    ].eval_with_dict(params)
    # (count-per-sub-group)*n_subgroups
    assert f32mul == n*m*n_subgroups
    assert f64div == 2*n*m*n_subgroups  # TODO why?
    assert f64add == n*m*n_subgroups
    assert i32add == n*m*n_subgroups


def test_op_counter_special_ops():

    knl = lp.make_kernel(
            "{[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
                c[i, j, k] = (2*a[i,j,k])%(2+b[i,j,k]/3.0)
                e[i, k] = (1+g[i,k])**(1+h[i,k+1])+rsqrt(g[i,k])*sin(g[i,k])
                """
            ],
            name="special_ops", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(knl,
                                  {"a": np.float32, "b": np.float32,
                                       "g": np.float64, "h": np.float64})
    op_map = lp.get_op_map(knl, subgroup_size=SGS, count_redundant_work=True,
                           count_within_subscripts=True)
    n_workgroups = 1
    group_size = 1
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}
    f32mul = op_map[lp.Op(np.float32, "mul", CG.SUBGROUP,
        "special_ops")].eval_with_dict(params)
    f32div = op_map[lp.Op(np.float32, "div", CG.SUBGROUP,
        "special_ops")].eval_with_dict(params)
    f32add = op_map[lp.Op(np.float32, "add", CG.SUBGROUP,
        "special_ops")].eval_with_dict(params)
    f64pow = op_map[lp.Op(np.float64, "pow", CG.SUBGROUP,
        "special_ops")].eval_with_dict(params)
    f64add = op_map[lp.Op(np.dtype(np.float64), "add", CG.SUBGROUP, "special_ops")
                    ].eval_with_dict(params)
    i32add = op_map[lp.Op(np.dtype(np.int32), "add", CG.SUBGROUP, "special_ops")
                    ].eval_with_dict(params)
    f64rsq = op_map[lp.Op(np.dtype(np.float64), "func:rsqrt", CG.SUBGROUP,
        "special_ops")].eval_with_dict(params)
    f64sin = op_map[lp.Op(np.dtype(np.float64), "func:sin", CG.SUBGROUP,
        "special_ops")].eval_with_dict(params)
    # (count-per-sub-group)*n_subgroups
    assert f32div == 2*n*m*ell*n_subgroups
    assert f32mul == f32add == n*m*ell*n_subgroups
    assert f64add == 3*n*m*n_subgroups
    assert f64pow == i32add == f64rsq == f64sin == n*m*n_subgroups


def test_op_counter_bitwise():

    knl = lp.make_kernel(
            "{[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
                c[i, j, k] = (a[i,j,k] | 1) + (b[i,j,k] & 1)
                e[i, k] = (g[i,k] ^ k)*(~h[i,k+1]) + (g[i, k] << (h[i,k] >> k))
                """
            ],
            name="bitwise", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(
            knl, {
                "a": np.int32, "b": np.int32,
                "g": np.int64, "h": np.int64})

    op_map = lp.get_op_map(knl, subgroup_size=SGS, count_redundant_work=True,
                           count_within_subscripts=False)
    n_workgroups = 1
    group_size = 1
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}
    print(op_map)
    i32add = op_map[
            lp.Op(np.int32, "add", CG.SUBGROUP, "bitwise")
            ].eval_with_dict(params)
    i32bw = op_map[
            lp.Op(np.int32, "bw", CG.SUBGROUP, "bitwise")
            ].eval_with_dict(params)
    i64bw = op_map[
            lp.Op(np.dtype(np.int64), "bw", CG.SUBGROUP, "bitwise")
            ].eval_with_dict(params)
    i64mul = op_map[
            lp.Op(np.dtype(np.int64), "mul", CG.SUBGROUP, "bitwise")
            ].eval_with_dict(params)
    i64add = op_map[
            lp.Op(np.dtype(np.int64), "add", CG.SUBGROUP, "bitwise")
            ].eval_with_dict(params)
    i64shift = op_map[
            lp.Op(np.dtype(np.int64), "shift", CG.SUBGROUP, "bitwise")
            ].eval_with_dict(params)
    # (count-per-sub-group)*n_subgroups
    assert i32add == n*m*ell*n_subgroups
    assert i32bw == 2*n*m*ell*n_subgroups
    assert i64bw == 2*n*m*n_subgroups
    assert i64add == i64mul == n*m*n_subgroups
    assert i64shift == 2*n*m*n_subgroups


def test_op_counter_triangular_domain():

    knl = lp.make_kernel(
            "{[i,j]: 0<=i<n and 0<=j<m and i<j}",
            """
            a[i, j] = b[i,j] * 2
            """,
            name="bitwise", assumptions="n,m >= 1")

    knl = lp.add_and_infer_dtypes(knl,
            {"b": np.float64})

    expect_fallback = False
    import islpy as isl
    try:
        isl.BasicSet.card  # noqa: B018
    except AttributeError:
        expect_fallback = True
    else:
        expect_fallback = False

    op_map = lp.get_op_map(
                    knl,
                    subgroup_size=SGS,
                    count_redundant_work=True
                    )[lp.Op(np.float64, "mul", CG.SUBGROUP, "bitwise")]
    value_dict = {"m": 13, "n": 200}
    flops = op_map.eval_with_dict(value_dict)

    n_workgroups = 1
    group_size = 1
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group

    if expect_fallback:
        assert flops == 144*n_subgroups
    else:
        assert flops == 78*n_subgroups


def test_mem_access_counter_basic():

    knl = lp.make_kernel(
            "[n,m,ell] -> {[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
                c[i, j, k] = a[i,j,k]*b[i,j,k]/3.0+a[i,j,k]
                e[i, k] = g[i,k]*h[i,k+1]
                """
            ],
            name="basic", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(knl,
            {"a": np.float32, "b": np.float32, "g": np.float64, "h": np.float64})

    mem_map = lp.get_mem_access_map(knl, count_redundant_work=True,
                                    subgroup_size=SGS)

    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}

    n_workgroups = 1
    group_size = 1
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group

    f32l = mem_map[lp.MemAccess("global", np.float32,
                        lid_strides={}, gid_strides={},
                        direction="load", variable="a",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="basic")
                   ].eval_with_dict(params)
    f32l += mem_map[lp.MemAccess("global", np.float32,
                        lid_strides={}, gid_strides={},
                        direction="load", variable="b",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="basic")
                    ].eval_with_dict(params)
    f64l = mem_map[lp.MemAccess("global", np.float64,
                        lid_strides={}, gid_strides={},
                        direction="load", variable="g",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="basic")
                   ].eval_with_dict(params)
    f64l += mem_map[lp.MemAccess("global", np.float64,
                        lid_strides={}, gid_strides={},
                        direction="load", variable="h",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="basic")
                    ].eval_with_dict(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert f32l == (3*n*m*ell)*n_subgroups
    assert f64l == (2*n*m)*n_subgroups

    f32s = mem_map[lp.MemAccess("global", np.dtype(np.float32),
                        lid_strides={}, gid_strides={},
                        direction="store", variable="c",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="basic")
                   ].eval_with_dict(params)
    f64s = mem_map[lp.MemAccess("global", np.dtype(np.float64),
                        lid_strides={}, gid_strides={},
                        direction="store", variable="e",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="basic")
                   ].eval_with_dict(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert f32s == (n*m*ell)*n_subgroups
    assert f64s == (n*m)*n_subgroups


def test_mem_access_counter_reduction():

    knl = lp.make_kernel(
            "{[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                "c[i, j] = sum(k, a[i, k]*b[k, j])"
            ],
            name="matmul", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(knl, {"a": np.float32, "b": np.float32})

    mem_map = lp.get_mem_access_map(knl, count_redundant_work=True,
                                    subgroup_size=SGS)
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}

    n_workgroups = 1
    group_size = 1
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group

    f32l = mem_map[lp.MemAccess("global", np.float32,
                        lid_strides={}, gid_strides={},
                        direction="load", variable="a",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="matmul")
                   ].eval_with_dict(params)
    f32l += mem_map[lp.MemAccess("global", np.float32,
                        lid_strides={}, gid_strides={},
                        direction="load", variable="b",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="matmul")
                    ].eval_with_dict(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert f32l == (2*n*m*ell)*n_subgroups

    f32s = mem_map[lp.MemAccess("global", np.dtype(np.float32),
                        lid_strides={}, gid_strides={},
                        direction="store", variable="c",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="matmul")
                   ].eval_with_dict(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert f32s == (n*ell)*n_subgroups

    ld_bytes = mem_map.filter_by(mtype=["global"], direction=["load"]
                                 ).to_bytes().eval_and_sum(params)
    st_bytes = mem_map.filter_by(mtype=["global"], direction=["store"]
                                 ).to_bytes().eval_and_sum(params)
    assert ld_bytes == 4*f32l
    assert st_bytes == 4*f32s


def test_mem_access_counter_logic():

    knl = lp.make_kernel(
            "{[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
                e[i,k] = if(not(k<ell-2) and k>6 or k/2==ell,
                    g[i,k]*2,
                    g[i,k]+h[i,k]/2)
                """
            ],
            name="logic", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(knl, {"g": np.float32, "h": np.float64})

    mem_map = lp.get_mem_access_map(knl, count_redundant_work=True,
                                    subgroup_size=SGS)
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}

    n_workgroups = 1
    group_size = 1
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group

    reduced_map = mem_map.group_by("mtype", "dtype", "direction")

    f32_g_l = reduced_map[lp.MemAccess("global", to_loopy_type(np.float32),
                                       direction="load")
                          ].eval_with_dict(params)
    f64_g_l = reduced_map[lp.MemAccess("global", to_loopy_type(np.float64),
                                       direction="load")
                          ].eval_with_dict(params)
    f64_g_s = reduced_map[lp.MemAccess("global", to_loopy_type(np.float64),
                                       direction="store")
                          ].eval_with_dict(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert f32_g_l == (2*n*m)*n_subgroups
    assert f64_g_l == (n*m)*n_subgroups
    assert f64_g_s == (n*m)*n_subgroups


def test_mem_access_counter_special_ops():

    knl = lp.make_kernel(
            "{[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
                c[i, j, k] = (2*a[i,j,k])%(2+b[i,j,k]/3.0)
                e[i, k] = (1+g[i,k])**(1+h[i,k+1])
                """
            ],
            name="special_ops", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(knl, {"a": np.float32, "b": np.float32,
                                            "g": np.float64, "h": np.float64})

    mem_map = lp.get_mem_access_map(knl, count_redundant_work=True,
                                    subgroup_size=SGS)
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}

    n_workgroups = 1
    group_size = 1
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group

    f32 = mem_map[lp.MemAccess("global", np.float32,
                        lid_strides={}, gid_strides={},
                        direction="load", variable="a",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="special_ops")
                  ].eval_with_dict(params)
    f32 += mem_map[lp.MemAccess("global", np.float32,
                        lid_strides={}, gid_strides={},
                        direction="load", variable="b",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="special_ops")
                   ].eval_with_dict(params)
    f64 = mem_map[lp.MemAccess("global", np.dtype(np.float64),
                        lid_strides={}, gid_strides={},
                        direction="load", variable="g",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="special_ops")
                  ].eval_with_dict(params)
    f64 += mem_map[lp.MemAccess("global", np.dtype(np.float64),
                        lid_strides={}, gid_strides={},
                        direction="load", variable="h",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="special_ops")
                   ].eval_with_dict(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert f32 == (2*n*m*ell)*n_subgroups
    assert f64 == (2*n*m)*n_subgroups

    f32 = mem_map[lp.MemAccess("global", np.float32,
                        lid_strides={}, gid_strides={},
                        direction="store", variable="c",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="special_ops")
                  ].eval_with_dict(params)
    f64 = mem_map[lp.MemAccess("global", np.float64,
                        lid_strides={}, gid_strides={},
                        direction="store", variable="e",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="special_ops")
                  ].eval_with_dict(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert f32 == (n*m*ell)*n_subgroups
    assert f64 == (n*m)*n_subgroups

    filtered_map = mem_map.filter_by(direction=["load"], variable=["a", "g"],
                         count_granularity=CG.SUBGROUP)
    tot = filtered_map.eval_and_sum(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert tot == (n*m*ell + n*m)*n_subgroups


def test_mem_access_counter_bitwise():

    knl = lp.make_kernel(
            "{[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
                c[i, j, k] = (a[i,j,k] | 1) + (b[i,j,k] & 1)
                e[i, k] = (g[i,k] ^ k)*(~h[i,k+1]) + (g[i, k] << (h[i,k] >> k))
                """
            ],
            name="bitwise", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(
            knl, {
                "a": np.int32, "b": np.int32,
                "g": np.int32, "h": np.int32})

    mem_map = lp.get_mem_access_map(knl, count_redundant_work=True,
                                    subgroup_size=SGS)
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}

    n_workgroups = 1
    group_size = 1
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group

    i32 = mem_map[lp.MemAccess("global", np.int32,
                        lid_strides={}, gid_strides={},
                        direction="load", variable="a",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="bitwise")
                  ].eval_with_dict(params)
    i32 += mem_map[lp.MemAccess("global", np.int32,
                        lid_strides={}, gid_strides={},
                        direction="load", variable="b",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="bitwise")
                   ].eval_with_dict(params)
    i32 += mem_map[lp.MemAccess("global", np.int32,
                        lid_strides={}, gid_strides={},
                        direction="load", variable="g",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="bitwise")
                   ].eval_with_dict(params)
    i32 += mem_map[lp.MemAccess("global", np.dtype(np.int32),
                        lid_strides={}, gid_strides={},
                        direction="load", variable="h",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="bitwise")
                   ].eval_with_dict(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert i32 == (4*n*m+2*n*m*ell)*n_subgroups

    i32 = mem_map[lp.MemAccess("global", np.int32,
                        lid_strides={}, gid_strides={},
                        direction="store", variable="c",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="bitwise")
                  ].eval_with_dict(params)
    i32 += mem_map[lp.MemAccess("global", np.int32,
                        lid_strides={}, gid_strides={},
                        direction="store", variable="e",
                        count_granularity=CG.SUBGROUP,
                        kernel_name="bitwise")
                   ].eval_with_dict(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert i32 == (n*m+n*m*ell)*n_subgroups


def test_mem_access_counter_mixed():
    knl = lp.make_kernel(
            "[n,m,ell] -> {[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
            c[i, j, k] = a[i,j,k]*b[i,j,k]/3.0+a[i,j,k]+x[i,k]
            e[i, k] = g[i,k]*(2+h[i,k])
            """
            ],
            name="mixed", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(knl, {
                "a": np.float32, "b": np.float32, "g": np.float64, "h": np.float64,
                "x": np.float32})

    group_size_0 = 65

    knl = lp.split_iname(knl, "j", group_size_0)
    knl = lp.tag_inames(knl, {"j_inner": "l.0", "j_outer": "g.0"})

    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}

    n_workgroups = div_ceil(ell, group_size_0)
    group_size = group_size_0
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group

    mem_map = lp.get_mem_access_map(knl, count_redundant_work=True,
                                    subgroup_size=SGS)
    f64uniform = mem_map[lp.MemAccess("global", np.float64,
                                lid_strides={}, gid_strides={},
                                direction="load", variable="g",
                                count_granularity=CG.SUBGROUP,
                                kernel_name="mixed")
                         ].eval_with_dict(params)
    f64uniform += mem_map[lp.MemAccess("global", np.float64,
                                lid_strides={}, gid_strides={},
                                direction="load", variable="h",
                                count_granularity=CG.SUBGROUP,
                                kernel_name="mixed")
                          ].eval_with_dict(params)
    f32uniform = mem_map[lp.MemAccess("global", np.float32,
                                lid_strides={}, gid_strides={},
                                direction="load", variable="x",
                                count_granularity=CG.SUBGROUP,
                                kernel_name="mixed")
                         ].eval_with_dict(params)
    f32nonconsec = mem_map[lp.MemAccess("global", np.dtype(np.float32),
                                lid_strides={0: Variable("m")},
                                gid_strides={0: Variable("m")*group_size_0},
                                direction="load",
                                variable="a",
                                count_granularity=CG.WORKITEM,
                                kernel_name="mixed")
                           ].eval_with_dict(params)
    f32nonconsec += mem_map[lp.MemAccess("global", np.dtype(np.float32),
                                lid_strides={0: Variable("m")},
                                gid_strides={0: Variable("m")*group_size_0},
                                direction="load",
                                variable="b",
                                count_granularity=CG.WORKITEM,
                                kernel_name="mixed")
                            ].eval_with_dict(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert f64uniform == (2*n*m)*n_subgroups
    assert f32uniform == (m*n)*n_subgroups

    expect_fallback = False
    import islpy as isl
    try:
        isl.BasicSet.card  # noqa: B018
    except AttributeError:
        expect_fallback = True
    else:
        expect_fallback = False

    if expect_fallback:
        if ell < group_size_0:
            assert f32nonconsec == 3*n*m*ell*n_workgroups
        else:
            assert f32nonconsec == 3*n*m*n_workgroups*group_size_0
    else:
        assert f32nonconsec == 3*n*m*ell

    f64uniform = mem_map[lp.MemAccess("global", np.float64,
                                lid_strides={}, gid_strides={},
                                direction="store", variable="e",
                                count_granularity=CG.SUBGROUP,
                                kernel_name="mixed")
                         ].eval_with_dict(params)
    f32nonconsec = mem_map[lp.MemAccess("global", np.float32,
                                lid_strides={0: Variable("m")},
                                gid_strides={0: Variable("m")*group_size_0},
                                direction="store",
                                variable="c",
                                count_granularity=CG.WORKITEM,
                                kernel_name="mixed")
                           ].eval_with_dict(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert f64uniform == m*n*n_subgroups

    if expect_fallback:
        if ell < group_size_0:
            assert f32nonconsec == n*m*ell*n_workgroups
        else:
            assert f32nonconsec == n*m*n_workgroups*group_size_0
    else:
        assert f32nonconsec == n*m*ell


def test_mem_access_counter_nonconsec():

    knl = lp.make_kernel(
            "[n,m,ell] -> {[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
            c[i, j, k] = a[i,j,k]*b[i,j,k]/3.0+a[i,j,k]
            e[i, k] = g[i,k]*(2+h[i,k])
            """
            ],
            name="non_consec", assumptions="n,m,ell >= 1")
    knl = lp.add_and_infer_dtypes(knl, {
                "a": np.float32, "b": np.float32, "g": np.float64, "h": np.float64})
    lsize0 = 16
    knl = lp.split_iname(knl, "i", lsize0)
    knl = lp.tag_inames(knl, {"i_inner": "l.0", "i_outer": "g.0"})

    mem_map = lp.get_mem_access_map(knl, count_redundant_work=True,
                                    subgroup_size=SGS)
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}
    f64nonconsec = mem_map[lp.MemAccess("global", np.float64,
                                lid_strides={0: Variable("m")},
                                gid_strides={0: Variable("m")*lsize0},
                                direction="load",
                                variable="g",
                                count_granularity=CG.WORKITEM,
                                kernel_name="non_consec")
                           ].eval_with_dict(params)
    f64nonconsec += mem_map[lp.MemAccess("global", np.float64,
                                lid_strides={0: Variable("m")},
                                gid_strides={0: Variable("m")*lsize0},
                                direction="load",
                                variable="h",
                                count_granularity=CG.WORKITEM,
                                kernel_name="non_consec")
                            ].eval_with_dict(params)
    f32nonconsec = mem_map[lp.MemAccess(
                            "global", np.dtype(np.float32),
                            lid_strides={0: Variable("m")*Variable("ell")},
                            gid_strides={0: Variable("m")*Variable("ell")*lsize0},
                            direction="load", variable="a",
                            count_granularity=CG.WORKITEM,
                            kernel_name="non_consec")
                           ].eval_with_dict(params)
    f32nonconsec += mem_map[lp.MemAccess(
                            "global", np.dtype(np.float32),
                            lid_strides={0: Variable("m")*Variable("ell")},
                            gid_strides={0: Variable("m")*Variable("ell")*lsize0},
                            direction="load", variable="b",
                            count_granularity=CG.WORKITEM,
                            kernel_name="non_consec")
                            ].eval_with_dict(params)
    assert f64nonconsec == 2*n*m
    assert f32nonconsec == 3*n*m*ell

    f64nonconsec = mem_map[lp.MemAccess("global", np.float64,
                                lid_strides={0: Variable("m")},
                                gid_strides={0: Variable("m")*lsize0},
                                direction="store",
                                variable="e",
                                count_granularity=CG.WORKITEM,
                                kernel_name="non_consec")
                           ].eval_with_dict(params)
    f32nonconsec = mem_map[lp.MemAccess(
                            "global", np.float32,
                            lid_strides={0: Variable("m")*Variable("ell")},
                            gid_strides={0: Variable("m")*Variable("ell")*lsize0},
                            direction="store", variable="c",
                            count_granularity=CG.WORKITEM,
                            kernel_name="non_consec")
                           ].eval_with_dict(params)
    assert f64nonconsec == n*m
    assert f32nonconsec == n*m*ell

    mem_map64 = lp.get_mem_access_map(knl, count_redundant_work=True,
                                      subgroup_size=64)
    f64nonconsec = mem_map64[lp.MemAccess(
                    "global",
                    np.float64,
                    lid_strides={0: Variable("m")},
                    gid_strides={0: Variable("m")*lsize0},
                    direction="load", variable="g",
                    count_granularity=CG.WORKITEM,
                    kernel_name="non_consec")
                    ].eval_with_dict(params)
    f64nonconsec += mem_map64[lp.MemAccess(
                    "global",
                    np.float64,
                    lid_strides={0: Variable("m")},
                    gid_strides={0: Variable("m")*lsize0},
                    direction="load", variable="h",
                    count_granularity=CG.WORKITEM,
                    kernel_name="non_consec")
                    ].eval_with_dict(params)
    f32nonconsec = mem_map64[lp.MemAccess(
                    "global",
                    np.dtype(np.float32),
                    lid_strides={0: Variable("m")*Variable("ell")},
                    gid_strides={0: Variable("m")*Variable("ell")*lsize0},
                    direction="load",
                    variable="a",
                    count_granularity=CG.WORKITEM,
                    kernel_name="non_consec")
                    ].eval_with_dict(params)
    f32nonconsec += mem_map64[lp.MemAccess(
                    "global",
                    np.dtype(np.float32),
                    lid_strides={0: Variable("m")*Variable("ell")},
                    gid_strides={0: Variable("m")*Variable("ell")*lsize0},
                    direction="load",
                    variable="b",
                    count_granularity=CG.WORKITEM,
                    kernel_name="non_consec")
                    ].eval_with_dict(params)
    assert f64nonconsec == 2*n*m
    assert f32nonconsec == 3*n*m*ell


def test_mem_access_counter_consec():

    knl = lp.make_kernel(
            "[n,m,ell] -> {[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
            c[i, j, k] = a[i,j,k]*b[i,j,k]/3.0+a[i,j,k]
            e[i, k] = g[i,k]*(2+h[i,k])
            """
            ],
            name="consec", assumptions="n,m,ell >= 1")
    knl = lp.add_and_infer_dtypes(knl, {
                "a": np.float32, "b": np.float32, "g": np.float64, "h": np.float64})
    knl = lp.tag_inames(knl, {"k": "l.0", "i": "g.0", "j": "g.1"})

    mem_map = lp.get_mem_access_map(knl, count_redundant_work=True,
                                    subgroup_size="guess")
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}

    f64consec = mem_map[lp.MemAccess(
                    "global", np.float64,
                    lid_strides={0: 1}, gid_strides={0: Variable("m")},
                    direction="load", variable="g",
                    count_granularity=CG.WORKITEM,
                    kernel_name="consec")
                    ].eval_with_dict(params)
    f64consec += mem_map[lp.MemAccess(
                    "global", np.float64,
                    lid_strides={0: 1}, gid_strides={0: Variable("m")},
                    direction="load", variable="h",
                    count_granularity=CG.WORKITEM,
                    kernel_name="consec")
                    ].eval_with_dict(params)
    f32consec = mem_map[lp.MemAccess(
                    "global", np.float32,
                    lid_strides={0: 1},
                    gid_strides={0: Variable("m")*Variable("ell"), 1: Variable("m")},
                    direction="load", variable="a",
                    count_granularity=CG.WORKITEM,
                    kernel_name="consec")
                    ].eval_with_dict(params)
    f32consec += mem_map[lp.MemAccess(
                    "global", np.dtype(np.float32),
                    lid_strides={0: 1},
                    gid_strides={0: Variable("m")*Variable("ell"), 1: Variable("m")},
                    direction="load", variable="b",
                    count_granularity=CG.WORKITEM,
                    kernel_name="consec")
                    ].eval_with_dict(params)
    assert f64consec == 2*n*m*ell
    assert f32consec == 3*n*m*ell

    f64consec = mem_map[lp.MemAccess(
                    "global", np.float64,
                    lid_strides={0: 1}, gid_strides={0: Variable("m")},
                    direction="store", variable="e",
                    count_granularity=CG.WORKITEM,
                    kernel_name="consec")
                    ].eval_with_dict(params)
    f32consec = mem_map[lp.MemAccess(
                    "global", np.float32,
                    lid_strides={0: 1},
                    gid_strides={0: Variable("m")*Variable("ell"), 1: Variable("m")},
                    direction="store", variable="c",
                    count_granularity=CG.WORKITEM,
                    kernel_name="consec")
                    ].eval_with_dict(params)
    assert f64consec == n*m*ell
    assert f32consec == n*m*ell


def test_mem_access_counter_global_temps():

    knl = lp.make_kernel(
            "[n,m,ell] -> {[i,j,k]: 0<=i<n and 0<=j<m and 0<=k<ell}",
            """
            <>a[i, j, k] = 3.1
            <>b[i, j] = 3.2
            """,
            assumptions="n,m,ell >= 1")
    knl = lp.add_and_infer_dtypes(knl, {"a,b": np.float32})

    # Change temporary b address space
    knl = lp.privatize_temporaries_with_inames(knl, "i,j", "b")
    knl = lp.set_temporary_address_space(knl, "b", "global")

    mem_map = lp.get_mem_access_map(knl, count_redundant_work=True,
                                    subgroup_size="guess")
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}

    # Count global accesses
    global_accesses = mem_map.filter_by(
        mtype=["global"]).sum().eval_with_dict(params)

    assert global_accesses == n*m


def test_count_granularity_val_checks():

    try:
        lp.MemAccess(count_granularity=CG.WORKITEM)
        lp.MemAccess(count_granularity=CG.SUBGROUP)
        lp.MemAccess(count_granularity=CG.WORKGROUP)
        lp.MemAccess(count_granularity=None)
        assert True
        lp.MemAccess(count_granularity="bushel")
        raise AssertionError()
    except ValueError:
        assert True

    try:
        lp.Op(count_granularity=CG.WORKITEM)
        lp.Op(count_granularity=CG.SUBGROUP)
        lp.Op(count_granularity=CG.WORKGROUP)
        lp.Op(count_granularity=None)
        assert True
        lp.Op(count_granularity="bushel")
        raise AssertionError()
    except ValueError:
        assert True


def test_barrier_counter_nobarriers():

    knl = lp.make_kernel(
            "[n,m,ell] -> {[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
                c[i, j, k] = a[i,j,k]*b[i,j,k]/3.0+a[i,j,k]
                e[i, k] = g[i,k]*h[i,k+1]
                """
            ],
            name="basic", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(knl, {"a": np.float32, "b": np.float32,
                                            "g": np.float64, "h": np.float64})
    sync_map = lp.get_synchronization_map(knl)
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}
    assert len(sync_map) == 1
    assert sync_map.filter_by(kind="kernel_launch").eval_and_sum(params) == 1


def test_barrier_counter_barriers():
    knl = lp.make_kernel(
            "[n,m,ell] -> {[i,k,j]: 0<=i<50 and 1<=k<98 and 0<=j<10}",
            [
                """
            c[i,j,k] = 2*a[i,j,k] {id=first}
            e[i,j,k] = c[i,j,k+1]+c[i,j,k-1] {dep=first}
            """
            ], [
                lp.TemporaryVariable("c", shape=(50, 10, 99)),
                ...
            ],
            name="weird2",
            )
    knl = lp.add_and_infer_dtypes(knl, {"a": np.int32})
    knl = lp.split_iname(knl, "k", 128, inner_tag="l.0")
    sync_map = lp.get_synchronization_map(knl)
    print(sync_map)
    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}
    barrier_count = sync_map.filter_by(kind="barrier_local").eval_and_sum(params)
    assert barrier_count == 50*10*2


def test_barrier_count_single():
    knl = lp.make_kernel(
            "{[i]: 0<=i<128}",
            """
            <> c[i] = 15*i {id=yoink}
            c[i+1] = c[i]  {dep=yoink}
            """)

    knl = lp.tag_inames(knl, {"i": "l.0"})
    sync_map = lp.get_synchronization_map(knl)
    print(sync_map)
    barrier_count = sync_map.filter_by(kind="barrier_local").eval_and_sum()
    assert barrier_count == 1


def test_all_counters_parallel_matmul():
    bsize = 16
    knl = lp.make_kernel(
            "{[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                "c[i, j] = sum(k, a[i, k]*b[k, j])"
            ],
            name="matmul", assumptions="n,m,ell >= 1")
    knl = lp.add_and_infer_dtypes(knl, {"a": np.float32, "b": np.float32})
    knl = lp.split_iname(knl, "i", bsize, outer_tag="g.0", inner_tag="l.1")
    knl = lp.split_iname(knl, "j", bsize, outer_tag="g.1", inner_tag="l.0")
    knl = lp.split_iname(knl, "k", bsize)
    knl = lp.add_prefetch(knl, "a", ["k_inner", "i_inner"], default_tag="l.auto")
    knl = lp.add_prefetch(knl, "b", ["j_inner", "k_inner"], default_tag="l.auto")

    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}
    group_size = bsize*bsize
    n_workgroups = div_ceil(n, bsize)*div_ceil(ell, bsize)
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group

    sync_map = lp.get_synchronization_map(knl)
    assert len(sync_map) == 2
    assert sync_map.filter_by(kind="kernel_launch").eval_and_sum(params) == 1
    assert sync_map.filter_by(kind="barrier_local").eval_and_sum(params) == 2*m/bsize

    op_map = lp.get_op_map(knl, subgroup_size=SGS, count_redundant_work=True)
    f32mul = op_map[
                        lp.Op(np.float32, "mul", CG.SUBGROUP, "matmul")
                        ].eval_with_dict(params)
    f32add = op_map[
                        lp.Op(np.float32, "add", CG.SUBGROUP, "matmul")
                        ].eval_with_dict(params)
    i32ops = op_map[
                        lp.Op(np.int32, "add", CG.SUBGROUP, "matmul")
                        ].eval_with_dict(params)
    i32ops += op_map[
                        lp.Op(np.dtype(np.int32), "mul", CG.SUBGROUP, "matmul")
                        ].eval_with_dict(params)

    # (count-per-sub-group)*n_subgroups
    assert f32mul+f32add == m*2*n_subgroups

    mem_access_map = lp.get_mem_access_map(knl, count_redundant_work=True,
                                           subgroup_size=SGS)

    f32s1lb = mem_access_map[lp.MemAccess("global", np.float32,
                             lid_strides={0: 1, 1: Variable("ell")},
                             gid_strides={1: bsize},
                             direction="load", variable="b",
                             count_granularity=CG.WORKITEM,
                             kernel_name="matmul")
                             ].eval_with_dict(params)
    f32s1la = mem_access_map[lp.MemAccess("global", np.float32,
                             lid_strides={0: 1, 1: Variable("m")},
                             gid_strides={0: Variable("m")*bsize},
                             direction="load",
                             variable="a", count_granularity=CG.WORKITEM,
                             kernel_name="matmul")
                             ].eval_with_dict(params)

    assert f32s1lb == n*m*ell/bsize
    assert f32s1la == n*m*ell/bsize

    f32coal = mem_access_map[lp.MemAccess("global", np.float32,
                             lid_strides={0: 1, 1: Variable("ell")},
                             gid_strides={0: Variable("ell")*bsize, 1: bsize},
                             direction="store", variable="c",
                             count_granularity=CG.WORKITEM,
                             kernel_name="matmul")
                             ].eval_with_dict(params)

    assert f32coal == n*ell

    local_mem_map = lp.get_mem_access_map(knl,
                        count_redundant_work=True,
                        subgroup_size=SGS).filter_by(mtype=["local"])

    local_mem_l = local_mem_map.filter_by(direction=["load"]
                                          ).eval_and_sum(params)
    # (count-per-sub-group)*n_subgroups
    assert local_mem_l == m*2*n_subgroups

    local_mem_l_a = local_mem_map[lp.MemAccess("local", np.dtype(np.float32),
                                               direction="load",
                                               lid_strides={1: 16},
                                               gid_strides={},
                                               variable="a_fetch",
                                               count_granularity=CG.SUBGROUP,
                                               kernel_name="matmul")
                                  ].eval_with_dict(params)
    local_mem_l_b = local_mem_map[lp.MemAccess("local", np.dtype(np.float32),
                                               direction="load",
                                               lid_strides={0: 1},
                                               gid_strides={},
                                               variable="b_fetch",
                                               count_granularity=CG.SUBGROUP,
                                               kernel_name="matmul")
                                  ].eval_with_dict(params)

    # (count-per-sub-group)*n_subgroups
    assert local_mem_l_a == local_mem_l_b == m*n_subgroups

    local_mem_s = local_mem_map.filter_by(direction=["store"]
                                          ).eval_and_sum(params)

    # (count-per-sub-group)*n_subgroups
    assert local_mem_s == m*2/bsize*n_subgroups


def test_floor_div_coefficient_collector():
    bsize = 16

    # kernel that shuffles local mem
    knl = lp.make_kernel(
        "{[i_outer,j_outer,i_inner,j_inner,r]: "
        "0<=i_outer<n/%d and 0<=j_outer<n/%d and "
        "0<=i_inner<%d and 0<=j_inner<%d and "
        "0<=r<rept}"
        % (bsize, bsize, bsize, bsize),
        [
            "for i_outer",
            "for j_outer",
            "<> loc[i_inner,j_inner] = 3.14f  {id=loc_init}",
            "loc[i_inner,(j_inner+r+4) %% %d] = loc[i_inner,(j_inner+r) %% %d]"
            "  {id=add,dep=loc_init}" % (bsize, bsize),
            "out0[i_outer*16+i_inner,j_outer*16+j_inner] = loc[i_inner,j_inner]"
            "  {id=store,dep=add}",
            "end",
            "end",
        ],
        name="local",
        lang_version=(2018, 2))

    knl = lp.add_and_infer_dtypes(knl, {"out0": np.float32})
    knl = lp.tag_inames(knl, "i_outer:g.1,i_inner:l.1,j_outer:g.0,j_inner:l.0")

    n = 512
    rept = 64
    params = {"n": n, "rept": rept}
    group_size = bsize*bsize
    n_workgroups = div_ceil(n, bsize)*div_ceil(n, bsize)
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group

    # count local f32 accesses
    m = lp.get_mem_access_map(knl, count_redundant_work=True, subgroup_size=SGS)
    f32_local = m.filter_by(dtype=[np.float32], mtype=["local"]).eval_and_sum(params)

    # (count-per-sub-group)*n_subgroups
    assert f32_local == 2*(rept+1)*n_subgroups


def test_mem_access_tagged_variables():
    bsize = 16
    knl = lp.make_kernel(
            "{[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                "c$mmresult[i, j] = sum(k, a$mmaload[i, k]*b$mmbload[k, j])"
            ],
            name="matmul", assumptions="n,m,ell >= 1")
    knl = lp.add_and_infer_dtypes(knl, {"a": np.float32, "b": np.float32})
    knl = lp.split_iname(knl, "i", bsize, outer_tag="g.0", inner_tag="l.1")
    knl = lp.split_iname(knl, "j", bsize, outer_tag="g.1", inner_tag="l.0")
    knl = lp.split_iname(knl, "k", bsize)
    # knl = lp.add_prefetch(knl, "a", ["k_inner", "i_inner"], default_tag="l.auto")
    # knl = lp.add_prefetch(knl, "b", ["j_inner", "k_inner"], default_tag="l.auto")

    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}
    group_size = bsize*bsize
    n_workgroups = div_ceil(n, bsize)*div_ceil(ell, bsize)
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group

    mem_access_map = lp.get_mem_access_map(knl, count_redundant_work=True,
                                           subgroup_size=SGS)

    f32s1lb = mem_access_map[
            lp.MemAccess("global", np.float32,
                lid_strides={0: 1},
                gid_strides={1: bsize},
                direction="load", variable="b",
                variable_tags=frozenset([lp.LegacyStringInstructionTag("mmbload")]),
                count_granularity=CG.WORKITEM,
                kernel_name="matmul")
            ].eval_with_dict(params)
    f32s1la = mem_access_map[
            lp.MemAccess("global", np.float32,
                lid_strides={1: Variable("m")},
                gid_strides={0: Variable("m")*bsize},
                direction="load",
                variable="a",
                variable_tags=frozenset([lp.LegacyStringInstructionTag("mmaload")]),
                count_granularity=CG.SUBGROUP,
                kernel_name="matmul")
            ].eval_with_dict(params)

    assert f32s1lb == n*m*ell

    # uniform: (count-per-sub-group)*n_subgroups
    assert f32s1la == m*n_subgroups

    f32coal = mem_access_map[
            lp.MemAccess("global", np.float32,
                lid_strides={0: 1, 1: Variable("ell")},
                gid_strides={0: Variable("ell")*bsize, 1: bsize},
                direction="store", variable="c",
                variable_tags=frozenset([lp.LegacyStringInstructionTag("mmresult")]),
                count_granularity=CG.WORKITEM,
                kernel_name="matmul")
            ].eval_with_dict(params)

    assert f32coal == n*ell


def test_gather_access_footprint():
    knl = lp.make_kernel(
            "{[i,k,j]: 0<=i,j,k<n}",
            [
                "c[i, j] = sum(k, a[i, k]*b[k, j]) + a[i,j]"
            ],
            name="matmul", assumptions="n >= 1")
    knl = lp.add_and_infer_dtypes(knl, {"a": np.float32, "b": np.float32})

    from loopy.statistics import count, gather_access_footprints
    fp = gather_access_footprints(knl)

    for key, footprint in fp.items():
        print(key, count(knl, footprint))


def test_gather_access_footprint_2():
    knl = lp.make_kernel(
            "{[i]: 0<=i<n}",
            "c[2*i] = a[i]",
            name="matmul", assumptions="n >= 1")
    knl = lp.add_and_infer_dtypes(knl, {"a": np.float32})

    from loopy.statistics import count, gather_access_footprints
    fp = gather_access_footprints(knl)

    params = {"n": 200}
    for key, footprint in fp.items():
        assert count(knl, footprint).eval_with_dict(params) == 200
        print(key, count(knl, footprint))


def test_summations_and_filters():

    knl = lp.make_kernel(
            "[n,m,ell] -> {[i,k,j]: 0<=i<n and 0<=k<m and 0<=j<ell}",
            [
                """
                c[i, j, k] = a[i,j,k]*b[i,j,k]/3.0+a[i,j,k]
                e[i, k+1] = -g[i,k]*h[i,k+1]
                """
            ],
            name="basic", assumptions="n,m,ell >= 1")

    knl = lp.add_and_infer_dtypes(knl,
            {"a": np.float32, "b": np.float32, "g": np.float64, "h": np.float64})

    n = 512
    m = 256
    ell = 128
    params = {"n": n, "m": m, "ell": ell}

    n_workgroups = 1
    group_size = 1
    subgroups_per_group = div_ceil(group_size, SGS)
    n_subgroups = n_workgroups*subgroups_per_group

    mem_map = lp.get_mem_access_map(knl, count_redundant_work=True,
                                    subgroup_size=SGS)

    loads_a = mem_map.filter_by(direction=["load"], variable=["a"],
                                count_granularity=[CG.SUBGROUP]
                                ).eval_and_sum(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert loads_a == (2*n*m*ell)*n_subgroups

    global_stores = mem_map.filter_by(mtype=["global"], direction=["store"],
                                      count_granularity=[CG.SUBGROUP]
                                      ).eval_and_sum(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert global_stores == (n*m*ell + n*m)*n_subgroups

    ld_bytes = mem_map.filter_by(mtype=["global"], direction=["load"],
                                 count_granularity=[CG.SUBGROUP]
                                 ).to_bytes().eval_and_sum(params)
    st_bytes = mem_map.filter_by(mtype=["global"], direction=["store"],
                                 count_granularity=[CG.SUBGROUP]
                                 ).to_bytes().eval_and_sum(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert ld_bytes == (4*n*m*ell*3 + 8*n*m*2)*n_subgroups
    assert st_bytes == (4*n*m*ell + 8*n*m)*n_subgroups

    # ignore stride and variable names in this map
    reduced_map = mem_map.group_by("mtype", "dtype", "direction")
    f32lall = reduced_map[lp.MemAccess("global", np.float32, direction="load")
                          ].eval_with_dict(params)
    f64lall = reduced_map[lp.MemAccess("global", np.float64, direction="load")
                          ].eval_with_dict(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert f32lall == (3*n*m*ell)*n_subgroups
    assert f64lall == (2*n*m)*n_subgroups

    op_map = lp.get_op_map(knl, subgroup_size=SGS, count_redundant_work=True,
                           count_within_subscripts=True)
    # for k, v in op_map.items():
    #    print(type(k), "\n", k.name, k.dtype, type(k.dtype), " :\n", v)

    op_map_dtype = op_map.group_by("dtype")
    f32 = op_map_dtype[lp.Op(dtype=np.float32)].eval_with_dict(params)
    f64 = op_map_dtype[lp.Op(dtype=np.float64)].eval_with_dict(params)
    i32 = op_map_dtype[lp.Op(dtype=np.int32)].eval_with_dict(params)
    assert f32 == n*m*ell*3
    assert f64 == n*m
    assert i32 == n*m*2

    addsub_all = op_map.filter_by(name=["add", "sub"]).eval_and_sum(params)
    f32ops_all = op_map.filter_by(dtype=[np.float32]).eval_and_sum(params)
    assert addsub_all == n*m*ell + n*m*2
    assert f32ops_all == n*m*ell*3

    non_field = op_map.filter_by(xxx=[np.float32]).eval_and_sum(params)
    assert non_field == 0

    ops_nodtype = op_map.group_by("name")
    ops_noname = op_map.group_by("dtype")
    mul_all = ops_nodtype[lp.Op(name="mul")].eval_with_dict(params)
    f64ops_all = ops_noname[lp.Op(dtype=np.float64)].eval_with_dict(params)
    assert mul_all == n*m*ell + n*m
    assert f64ops_all == n*m

    def func_filter(key):
        return key.lid_strides == {} and key.dtype == to_loopy_type(np.float64) and \
               key.direction == "load"
    f64l = mem_map.filter_by_func(func_filter).eval_and_sum(params)

    # uniform: (count-per-sub-group)*n_subgroups
    assert f64l == (2*n*m)*n_subgroups


def test_strided_footprint():
    param_dict = {"n": 2**20}
    knl = lp.make_kernel(
        "[n] -> {[i]: 0<=i<n}",
        [
            "z[i] = x[3*i]"
        ], name="s3")

    knl = lp.add_and_infer_dtypes(knl, {"x": np.float32})

    unr = 4
    bx = 256

    knl = lp.split_iname(knl, "i", bx*unr, outer_tag="g.0", slabs=(0, 1))
    knl = lp.split_iname(knl, "i_inner", bx, outer_tag="unr", inner_tag="l.0")

    footprints = lp.gather_access_footprints(knl)
    x_l_foot = footprints["x", "read"]

    from loopy.statistics import count
    num = count(knl, x_l_foot).eval_with_dict(param_dict)
    denom = count(knl, x_l_foot.remove_divs()).eval_with_dict(param_dict)

    assert 2*num < denom


def test_stats_on_callable_kernel():
    callee = lp.make_function(
            "{[i, j]: 0<=i, j< 20}",
            """
            y[i] = sum(j, A[i,j]*x[j])
            """, name="matvec20x20")

    caller = lp.make_kernel(
            "{:}",
            """
            y[:]  = matvec20x20(A[:,:], x[:])
            """,
            [
                lp.GlobalArg("x,y", shape=(20,), dtype=np.float64),
                lp.GlobalArg("A", shape=(20, 20), dtype=np.float64),
                ],
            name="matvec")
    caller = lp.merge([caller, callee])

    op_map = lp.get_op_map(caller, subgroup_size=SGS, count_redundant_work=True,
                           count_within_subscripts=True)
    f64_add = op_map.filter_by(name="add").eval_and_sum({})
    assert f64_add == 400


def test_stats_on_callable_kernel_within_loop():
    callee = lp.make_function(
            "{[i, j]: 0<=i, j< 20}",
            """
            y[i] = sum(j, A[i,j]*x[j])
            """, name="matvec20x20")

    caller = lp.make_kernel(
            "{[i]: 0<=i< 20}",
            """
            y[i, :]  = matvec20x20(A[:,:], x[i, :])
            """,
            [
                lp.GlobalArg("x,y", shape=(20, 20), dtype=np.float64),
                lp.GlobalArg("A", shape=(20, 20), dtype=np.float64),
                ],
            name="matmat")
    caller = lp.merge([caller, callee])

    op_map = lp.get_op_map(caller, subgroup_size=SGS, count_redundant_work=True,
                           count_within_subscripts=True)

    f64_add = op_map.filter_by(name="add").eval_and_sum({})
    assert f64_add == 8000


def test_callable_kernel_with_substitution():
    callee = lp.make_function(
            "{[i, j]: 0<=i, j< n}",
            """
            y[i] = sum(j, A[i,j]*x[j])
            """,
            [lp.ValueArg("n"), ...],
            name="matvec")

    caller = lp.make_kernel(
            "{[i]: 0<=i< 20}",
            """
            y[i, :]  = matvec(20, A[:,:], x[i, :])
            """,
            [
                lp.GlobalArg("x,y", shape=(20, 20), dtype=np.float64),
                lp.GlobalArg("A", shape=(20, 20), dtype=np.float64),
                ],
            name="matmat")
    caller = lp.merge([caller, callee])

    op_map = lp.get_op_map(caller, subgroup_size=SGS, count_redundant_work=True,
                           count_within_subscripts=True)

    f64_add = op_map.filter_by(name="add").eval_and_sum({})
    assert f64_add == 8000


def test_no_loop_ops():
    # See https://github.com/inducer/loopy/issues/211

    knl = lp.make_kernel(
        "{ : }",
        """
        c = a + b
        d = 2*c + a + b
        """)

    knl = lp.add_dtypes(knl, {"a": np.float64, "b": np.float64})

    op_map = lp.get_op_map(knl, subgroup_size=SGS, count_redundant_work=True,
                           count_within_subscripts=True)
    f64_add = op_map.filter_by(name="add").eval_and_sum({})
    f64_mul = op_map.filter_by(name="mul").eval_and_sum({})
    assert f64_add == 3
    assert f64_mul == 1


def test_within_stats():
    import numpy as np
    import pytest

    import loopy as lp

    knl = lp.make_kernel(
        "{[i]: 0<=i<10}",
        """
        out1[i] = 2.0f * i  {tags=writes_float}
        out2[i] = 3 * i     {tags=writes_int}
        out3[i] = 7.0f * i  {tags=writes_float}
        """)

    op_map = lp.get_op_map(knl, subgroup_size="guess", within="tag:writes_float")

    ops_dtype = op_map.group_by("dtype")

    f32ops = ops_dtype[lp.Op(dtype=np.float32)].eval_with_dict({})

    assert f32ops == 20

    with pytest.raises(KeyError):
        _ = ops_dtype[lp.Op(dtype=np.int32)].eval_with_dict({})

    mem_map = lp.get_mem_access_map(knl, subgroup_size="guess",
                                    within="tag:writes_float")

    mem_dtype = mem_map.group_by("dtype")
    mem_ops = mem_dtype[lp.MemAccess(dtype=np.float32)].eval_with_dict({})
    assert mem_ops == 20

    with pytest.raises(KeyError):
        _ = ops_dtype[lp.MemAccess(dtype=np.int32)].eval_with_dict({})


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from dataclasses import dataclass, replace
from typing import TYPE_CHECKING, cast
from warnings import warn

import numpy as np
from constantdict import constantdict

from islpy import dim_type
from pytools import MovedFunctionDeprecationWrapper

from loopy.diagnostic import LoopyError
from loopy.kernel import LoopKernel
from loopy.kernel.data import AddressSpace, ImageArg, TemporaryVariable, auto
from loopy.kernel.function_interface import CallableKernel, ScalarCallable
from loopy.translation_unit import TranslationUnit, for_each_kernel
from loopy.types import LoopyType
from loopy.typing import assert_tuple


if TYPE_CHECKING:
    from pymbolic import ArithmeticExpression

    from loopy.typing import Expression


# {{{ convenience: add_prefetch

# {{{ process footprint_subscripts

def _add_kernel_axis(kernel, axis_name, start, stop, base_inames):
    from loopy.kernel.tools import DomainChanger
    domch = DomainChanger(kernel, base_inames)

    domain = domch.domain
    new_dim_idx = domain.dim(dim_type.set)
    domain = (domain
            .insert_dims(dim_type.set, new_dim_idx, 1)
            .set_dim_name(dim_type.set, new_dim_idx, axis_name))

    from loopy.symbolic import get_dependencies
    deps = get_dependencies(start) | get_dependencies(stop)
    assert deps <= kernel.all_params()

    param_names = domain.get_var_names(dim_type.param)
    for dep in deps:
        if dep not in param_names:
            new_dim_idx = domain.dim(dim_type.param)
            domain = (domain
                    .insert_dims(dim_type.param, new_dim_idx, 1)
                    .set_dim_name(dim_type.param, new_dim_idx, dep))

    from loopy.isl_helpers import make_slab
    slab = make_slab(domain.get_space(), axis_name, start, stop)

    domain = domain & slab

    return kernel.copy(domains=domch.get_domains_with(domain))


def _process_footprint_subscripts(kernel, rule_name, sweep_inames,
        footprint_subscripts, arg):
    """Track applied iname rewrites, deal with slice specifiers ':'."""

    name_gen = kernel.get_var_name_generator()

    from pymbolic.primitives import Variable

    if footprint_subscripts is None:
        return kernel, rule_name, sweep_inames, []

    if not isinstance(footprint_subscripts, (list, tuple)):
        footprint_subscripts = [footprint_subscripts]

    inames_to_be_removed = []

    new_footprint_subscripts = []
    for fsub in footprint_subscripts:
        if isinstance(fsub, str):
            from loopy.symbolic import parse
            fsub = parse(fsub)

        if not isinstance(fsub, tuple):
            fsub = (fsub,)

        if len(fsub) != arg.num_user_axes():
            raise ValueError("sweep index '%s' has the wrong number of dimensions"
                    % str(fsub))

        for subst_map in kernel.applied_iname_rewrites:
            from pymbolic.mapper.substitutor import make_subst_func

            from loopy.symbolic import SubstitutionMapper
            fsub = SubstitutionMapper(make_subst_func(subst_map))(fsub)

        from loopy.symbolic import get_dependencies
        fsub_dependencies = get_dependencies(fsub)

        new_fsub = []
        for axis_nr, fsub_axis in enumerate(fsub):
            from pymbolic.primitives import Slice
            if isinstance(fsub_axis, Slice):
                if fsub_axis.children != (None,):
                    raise NotImplementedError("add_prefetch only "
                            "supports full slices")

                axis_name = name_gen(
                        based_on="%s_fetch_axis_%d" % (arg.name, axis_nr))

                kernel = _add_kernel_axis(kernel, axis_name, 0, arg.shape[axis_nr],
                        frozenset(sweep_inames) | fsub_dependencies)
                sweep_inames = [*sweep_inames, axis_name]

                inames_to_be_removed.append(axis_name)
                new_fsub.append(Variable(axis_name))

            else:
                new_fsub.append(fsub_axis)

        new_footprint_subscripts.append(tuple(new_fsub))
        del new_fsub

    footprint_subscripts = new_footprint_subscripts
    del new_footprint_subscripts

    subst_use = [Variable(rule_name)(*si) for si in footprint_subscripts]
    return kernel, subst_use, sweep_inames, inames_to_be_removed

# }}}


def add_prefetch_for_single_kernel(kernel, callables_table, var_name,
        sweep_inames=None, dim_arg_names=None,

        default_tag=None,

        rule_name=None,
        temporary_name=None,
        temporary_address_space=None, temporary_scope=None,
        footprint_subscripts=None,
        fetch_bounding_box=False,
        fetch_outer_inames=None,
        prefetch_insn_id=None,
        within=None):
    """See :func:`add_prefetch` for detailed, user-facing documentation."""

    assert isinstance(kernel, LoopKernel)
    if sweep_inames is None:
        sweep_inames = []

    # {{{ fish indexing out of var_name and into footprint_subscripts

    from loopy.symbolic import parse
    parsed_var_name = parse(var_name)

    from pymbolic.primitives import Subscript, Variable
    if isinstance(parsed_var_name, Variable):
        # nothing to see
        pass
    elif isinstance(parsed_var_name, Subscript):
        if footprint_subscripts is not None:
            raise TypeError("if footprint_subscripts is specified, then var_name "
                    "may not contain a subscript")

        assert isinstance(parsed_var_name.aggregate, Variable)
        footprint_subscripts = [parsed_var_name.index]
        parsed_var_name = parsed_var_name.aggregate
    else:
        raise ValueError("var_name must either be a variable name or a subscript")

    # }}}

    # {{{ fish out tags

    from loopy.symbolic import TaggedVariable
    if isinstance(parsed_var_name, TaggedVariable):
        var_name = parsed_var_name.name
        tags = parsed_var_name.tags
    else:
        var_name = parsed_var_name.name
        tags = ()

    # }}}

    c_name = var_name
    from loopy.kernel.instruction import LegacyStringInstructionTag
    tag_suffix = "_".join(tag.value for tag in tags
            if isinstance(tag, LegacyStringInstructionTag))
    if tag_suffix:
        c_name = c_name + "_" + tag_suffix

    var_name_gen = kernel.get_var_name_generator()

    if rule_name is None:
        rule_name = var_name_gen("%s_fetch_rule" % c_name)
    if temporary_name is None:
        temporary_name = var_name_gen("%s_fetch" % c_name)

    var_descr = kernel.get_var_descriptor(var_name)

    # {{{ make parameter names and unification template

    parameters = []
    for i in range(var_descr.num_user_axes()):
        based_on = "%s_dim_%d" % (c_name, i)
        if var_descr.dim_names is not None:
            based_on = "{}_dim_{}".format(c_name, var_descr.dim_names[i])
        if dim_arg_names is not None and i < len(dim_arg_names):
            based_on = dim_arg_names[i]

        par_name = var_name_gen(based_on=based_on)
        parameters.append(par_name)

    from pymbolic import var
    uni_template = parsed_var_name
    if len(parameters) > 1:
        uni_template = uni_template[
                tuple(var(par_name) for par_name in parameters)]
    elif len(parameters) == 1:
        uni_template = uni_template[var(parameters[0])]

    # }}}

    from loopy.transform.subst import extract_subst
    kernel = extract_subst(kernel, rule_name, uni_template, parameters,
            within=within)

    if isinstance(sweep_inames, str):
        sweep_inames = [s.strip() for s in sweep_inames.split(",")]
    else:
        # copy, standardize to list
        sweep_inames = list(sweep_inames)

    kernel, subst_use, sweep_inames, inames_to_be_removed = \
            _process_footprint_subscripts(
                    kernel,  rule_name, sweep_inames,
                    footprint_subscripts, var_descr)

    # Our _not_provided is actually a different object from the one in the
    # precompute module, but precompute actually uses that to adjust its
    # warning message.

    from loopy.transform.precompute import precompute_for_single_kernel
    new_kernel = precompute_for_single_kernel(kernel, callables_table,
            subst_use, sweep_inames, precompute_inames=dim_arg_names,
            default_tag=default_tag, dtype=var_descr.dtype,
            fetch_bounding_box=fetch_bounding_box,
            temporary_name=temporary_name,
            temporary_address_space=temporary_address_space,
            precompute_outer_inames=fetch_outer_inames,
            compute_insn_id=prefetch_insn_id,
            within=within)

    # {{{ remove inames that were temporarily added by slice sweeps

    new_domains = new_kernel.domains[:]

    for iname in inames_to_be_removed:
        home_domain_index = kernel.get_home_domain_index(iname)
        domain = new_domains[home_domain_index]

        dt, idx = domain.get_var_dict()[iname]
        assert dt == dim_type.set

        new_domains[home_domain_index] = domain.project_out(dt, idx, 1)

    new_kernel = new_kernel.copy(domains=new_domains)

    # }}}

    # If the rule survived past precompute() (i.e. some accesses fell outside
    # the footprint), get rid of it before moving on.
    if rule_name in new_kernel.substitutions:
        from loopy.transform.subst import expand_subst
        return expand_subst(new_kernel, "... > id:"+rule_name)
    else:
        return new_kernel


def add_prefetch(t_unit,
                 var_name, sweep_inames=None, dim_arg_names=None,
                 default_tag=None,
                 rule_name=None, temporary_name=None,
                 temporary_address_space=None, temporary_scope=None,
                 footprint_subscripts=None, fetch_bounding_box=False,
                 fetch_outer_inames=None, prefetch_insn_id=None, within=None):
    """Prefetch all accesses to the variable *var_name*, with all accesses
    being swept through *sweep_inames*.

    :arg var_name: A string, the name of the variable being prefetched.
        This may be a 'tagged variable name' (such as ``field$mytag``
        to restrict the effect of the operation to only variable accesses
        with a matching tag.

        This may also be a subscripted version of the variable, in which
        case this access dictates the footprint that is prefetched,
        e.g. ``A[:,:]`` or ``field[i,j,:,:]``. In this case, accesses
        in the kernel are disregarded.

    :arg sweep_inames: A list of inames, or a comma-separated string of them.
        This routine 'sweeps' all accesses to *var_name* through all allowed
        values of the *sweep_inames* to generate a footprint. All values
        in this footprint are then stored in a temporary variable, and
        the original variable accesses replaced with accesses to this
        temporary.

    :arg dim_arg_names: List of names representing each fetch axis.
        These names show up as inames in the generated fetch code

    :arg default_tag: The :ref:`implementation tag <iname-tags>` to
        assign to the inames driving the prefetch code. Use *None* to
        leave them undefined (to assign them later by hand). The current
        default will make them local axes and automatically split them to
        fit the work group size, but this default will disappear in favor
        of simply leaving them untagged in 2019.x. For 2018.x, a warning
        will be issued if no *default_tag* is specified.

    :arg rule_name: base name of the generated temporary variable.
    :arg temporary_name: The name of the temporary to be used.
    :arg temporary_address_space: The :class:`AddressSpace` to use for the
        temporary.
    :arg footprint_subscripts: A list of tuples indicating the index (i.e.
        subscript) tuples used to generate the footprint.

        If only one such set of indices is desired, this may also be specified
        directly by putting an index expression into *var_name*. Substitutions
        such as those occurring in dimension splits are recorded and also
        applied to these indices.

    :arg fetch_bounding_box: To fit within :mod:`loopy`'s execution model,
        the 'footprint' of the fetch currently has to be a convex set.
        Sometimes this is not the case, e.g. for a high-order stencil::

              o
              o
            ooooo
              o
              o

        The footprint of the stencil when 'swept' over a base domain
        would look like this, and because of the 'missing corners',
        this set is not convex::

              oooooooooo
              oooooooooo
            oooooooooooooo
            oooooooooooooo
            oooooooooooooo
            oooooooooooooo
              oooooooooo
              oooooooooo

        Passing ``fetch_bounding_box=True`` gives :mod:`loopy` permission
        to instead fetch the 'bounding box' of the footprint, i.e.
        this set in the stencil example::

            OOooooooooooOO
            OOooooooooooOO
            oooooooooooooo
            oooooooooooooo
            oooooooooooooo
            oooooooooooooo
            OOooooooooooOO
            OOooooooooooOO

        Note the added corners marked with "``O``". The resulting footprint is
        guaranteed to be convex.


    :arg fetch_outer_inames: The inames within which the fetch
        instruction is nested. If *None*, make an educated guess.

    :arg prefetch_insn_id: The ID of the instruction generated to perform the
        prefetch.

    :arg within: a stack match as understood by
        :func:`loopy.match.parse_stack_match` to select the instructions where
        *var_name* is to be prefetched.

    This function internally uses :func:`extract_subst` and :func:`precompute`.
    """
    assert isinstance(t_unit, TranslationUnit)

    new_callables = {}
    for func_id, in_knl_callable in t_unit.callables_table.items():
        if isinstance(in_knl_callable, CallableKernel):
            new_subkernel = add_prefetch_for_single_kernel(
                    in_knl_callable.subkernel, t_unit.callables_table,
                    var_name=var_name,
                    sweep_inames=sweep_inames,
                    dim_arg_names=dim_arg_names,
                    default_tag=default_tag,
                    rule_name=rule_name,
                    temporary_name=temporary_name,
                    temporary_address_space=temporary_address_space,
                    temporary_scope=temporary_scope,
                    footprint_subscripts=footprint_subscripts,
                    fetch_bounding_box=fetch_bounding_box,
                    fetch_outer_inames=fetch_outer_inames,
                    prefetch_insn_id=prefetch_insn_id,
                    within=within)
            in_knl_callable = in_knl_callable.copy(
                    subkernel=new_subkernel)

        elif isinstance(in_knl_callable, ScalarCallable):
            pass
        else:
            raise NotImplementedError("Unknown type of callable %s." % (
                type(in_knl_callable).__name__))

        new_callables[func_id] = in_knl_callable

    return t_unit.copy(callables_table=constantdict(new_callables))

# }}}


# {{{ change variable kinds

@for_each_kernel
def change_arg_to_image(kernel, name):
    new_args = []
    for arg in kernel.args:
        if arg.name == name:
            assert arg.offset == 0
            assert arg.shape is not None
            new_args.append(ImageArg(arg.name, dtype=arg.dtype, shape=arg.shape))
        else:
            new_args.append(arg)

    return kernel.copy(args=new_args)

# }}}


# {{{ tag array axes

@for_each_kernel
def tag_array_axes(kernel, ary_names, dim_tags):
    """
    :arg dim_tags: a tuple of
        :class:`loopy.kernel.array.ArrayDimImplementationTag` or a string that
        parses to one. See :func:`loopy.kernel.array.parse_array_dim_tags` for a
        description of the allowed string format.

        For example, *dim_tags* could be ``"N2,N0,N1"`` to determine
        that the second axis is the fastest-varying, the last is
        the next-fastest, and the first is the slowest.

    .. versionchanged:: 2016.2

        This function was called ``tag_data_axes`` before version 2016.2.
    """

    from loopy.kernel.tools import ArrayChanger

    if isinstance(ary_names, str):
        ary_names = [ary_name.strip() for ary_name in ary_names.split(",")]

    for ary_name in ary_names:
        achng = ArrayChanger(kernel, ary_name)
        ary = achng.get()

        from loopy.kernel.array import parse_array_dim_tags
        new_dim_tags = parse_array_dim_tags(dim_tags,
                n_axes=ary.num_user_axes(),
                use_increasing_target_axes=ary.max_target_axes > 1,
                dim_names=ary.dim_names)

        ary = ary.copy(dim_tags=tuple(new_dim_tags))

        kernel = achng.with_changed_array(ary)

    return kernel


tag_data_axes = (
        MovedFunctionDeprecationWrapper(tag_array_axes))

# }}}


# {{{ set_array_axis_names

@for_each_kernel
def set_array_axis_names(kernel, ary_names, dim_names):
    """
    .. versionchanged:: 2016.2

        This function was called ``set_array_dim_names`` before version 2016.2.
    """
    from loopy.kernel.tools import ArrayChanger
    if isinstance(ary_names, str):
        ary_names = ary_names.split(",")

    if isinstance(dim_names, str):
        dim_names = tuple(dim_names.split(","))

    for ary_name in ary_names:
        achng = ArrayChanger(kernel, ary_name)
        ary = achng.get()

        ary = ary.copy(dim_names=dim_names)

        kernel = achng.with_changed_array(ary)

    return kernel


set_array_dim_names = (MovedFunctionDeprecationWrapper(
    set_array_axis_names))

# }}}


# {{{ remove_unused_arguments

@for_each_kernel
def remove_unused_arguments(kernel):
    new_args = []

    import loopy as lp
    exp_kernel = lp.expand_subst(kernel)

    refd_vars = set(kernel.all_params())
    for insn in exp_kernel.instructions:
        refd_vars.update(insn.dependency_names())

    from itertools import chain

    from loopy.kernel.array import ArrayBase, FixedStrideArrayDimTag
    from loopy.symbolic import get_dependencies

    def tolerant_get_deps(expr):
        if expr is None or expr is lp.auto:
            return set()
        return get_dependencies(expr)

    for ary in chain(kernel.args, kernel.temporary_variables.values()):
        if isinstance(ary, ArrayBase):
            refd_vars.update(
                    tolerant_get_deps(ary.shape)
                    | tolerant_get_deps(ary.offset))

            for dim_tag in ary.dim_tags:
                if isinstance(dim_tag, FixedStrideArrayDimTag):
                    refd_vars.update(
                            tolerant_get_deps(dim_tag.stride))

    for arg in kernel.args:
        if arg.name in refd_vars:
            new_args.append(arg)

    return kernel.copy(args=new_args)

# }}}


# {{{ alias_temporaries

@for_each_kernel
def alias_temporaries(kernel, names, base_name_prefix=None,
        synchronize_for_exclusive_use=True):
    """Sets all temporaries given by *names* to be backed by a single piece of
    storage.

    :arg synchronize_for_exclusive_use: A :class:`bool`. If ``True``, this also
        introduces ordering structures ("groups") to prevent the usage to ensure
        that the live ranges (i.e. the regions of code where each of the
        temporaries is used) do not overlap. This will allow two (or more)
        temporaries to share the same storage space as long as their live
        ranges do not need to be concurrent.
    :arg base_name_prefix: an identifier to be used for the common storage
        area

    .. versionchanged:: 2016.3

        Added *synchronize_for_exclusive_use* flag.
        ``synchronize_for_exclusive_use=True`` was the previous default
        behavior.
    """
    gng = kernel.get_group_name_generator()
    group_names = [gng("tmpgrp_"+name) for name in names]

    if base_name_prefix is None:
        base_name_prefix = "temp_storage"

    from pytools import UniqueNameGenerator
    vng = UniqueNameGenerator(
            kernel.all_variable_names()
            | {tv.base_storage
                for tv in kernel.temporary_variables.values()
                if tv.base_storage is not None})
    base_name = vng(base_name_prefix)

    names_set = set(names)

    if synchronize_for_exclusive_use:
        new_insns = []
        for insn in kernel.instructions:
            temp_deps = insn.dependency_names() & names_set

            if not temp_deps:
                new_insns.append(insn)
                continue

            if len(temp_deps) > 1:
                raise LoopyError("Instruction {insn} refers to multiple of the "
                        "temporaries being aliased, namely '{temps}'. Cannot alias."
                        .format(
                            insn=insn.id,
                            temps=", ".join(temp_deps)))

            temp_name, = temp_deps
            temp_idx = names.index(temp_name)
            group_name = group_names[temp_idx]
            other_group_names = (
                    frozenset(group_names[:temp_idx])
                    | frozenset(group_names[temp_idx+1:]))

            new_insns.append(
                    insn.copy(
                        groups=insn.groups | frozenset([group_name]),
                        conflicts_with_groups=(
                            insn.conflicts_with_groups | other_group_names)))
    else:
        new_insns = kernel.instructions

    new_temporary_variables = {}
    for tv in kernel.temporary_variables.values():
        if tv.name in names_set:
            if tv.base_storage is not None:
                raise LoopyError("temporary variable '{tv}' already has "
                        "a defined storage array -- cannot alias"
                        .format(tv=tv.name))

            new_temporary_variables[tv.name] = \
                    tv.copy(
                            base_storage=base_name,
                            _base_storage_access_may_be_aliasing=False)
        else:
            new_temporary_variables[tv.name] = tv

    return kernel.copy(
            instructions=new_insns,
            temporary_variables=new_temporary_variables)

# }}}


# {{{ set argument order

@for_each_kernel
def set_argument_order(kernel, arg_names):
    """
    :arg arg_names: A list (or comma-separated string) or argument
        names. All arguments must be in this list.
    """
    # FIXME: @inducer -- should this only affect the root kernel, or should it
    # take a within?

    if isinstance(arg_names, str):
        arg_names = arg_names.split(",")

    new_args = []
    old_arg_dict = kernel.arg_dict.copy()

    for arg_name in arg_names:
        try:
            arg = old_arg_dict.pop(arg_name)
        except KeyError as err:
            raise LoopyError("unknown argument '%s'" % arg_name) from err

        new_args.append(arg)

    if old_arg_dict:
        raise LoopyError("incomplete argument list passed "
                "to set_argument_order. Left over: '%s'"
                % ", ".join(arg_name for arg_name in old_arg_dict))

    return kernel.copy(args=new_args)

# }}}


# {{{ rename argument

@for_each_kernel
def rename_argument(kernel, old_name, new_name, existing_ok=False):
    """
    .. versionadded:: 2016.2
    """

    var_name_gen = kernel.get_var_name_generator()

    if old_name not in kernel.arg_dict:
        raise LoopyError("old arg name '%s' does not exist" % old_name)

    does_exist = var_name_gen.is_name_conflicting(new_name)

    if does_exist and not existing_ok:
        raise LoopyError("argument name '%s' conflicts with an existing identifier"
                "--cannot rename" % new_name)

    # {{{ instructions

    from pymbolic import var
    subst_dict = {old_name: var(new_name)}

    from pymbolic.mapper.substitutor import make_subst_func

    from loopy.symbolic import (
        RuleAwareSubstitutionMapper,
        SubstitutionRuleMappingContext,
    )
    rule_mapping_context = SubstitutionRuleMappingContext(
            kernel.substitutions, var_name_gen)
    smap = RuleAwareSubstitutionMapper(rule_mapping_context,
                    make_subst_func(subst_dict),
                    within=lambda kernel, insn, stack: True)

    kernel = rule_mapping_context.finish_kernel(smap.map_kernel(kernel))

    # }}}

    # {{{ args

    new_args = []
    for arg in kernel.args:
        if arg.name == old_name:
            arg = arg.copy(name=new_name)

        new_args.append(arg)

    # }}}

    # {{{ domain/assumptions

    def rename_arg_in_basic_set(dom):
        dom_var_dict = dom.get_var_dict()
        if old_name in dom_var_dict:
            dt, pos = dom_var_dict[old_name]
            dom = dom.set_dim_name(dt, pos, new_name)

        return dom

    new_domains = []
    for dom in kernel.domains:
        dom = rename_arg_in_basic_set(dom)
        new_domains.append(dom)

    new_assumptions = rename_arg_in_basic_set(kernel.assumptions)

    # }}}

    return kernel.copy(domains=new_domains, args=new_args,
            assumptions=new_assumptions)

# }}}


# {{{ set temporary address space

@for_each_kernel
def set_temporary_address_space(kernel, temp_var_names, address_space):
    """
    :arg temp_var_names: a container with membership checking,
        or a comma-separated string of variables for which the
        address space is to be set.
    :arg address_space: One of the values from :class:`loopy.AddressSpace`, or one
        of the strings ``"private"``, ``"local"``, or ``"global"``.
    """

    if isinstance(temp_var_names, str):
        temp_var_names = [s.strip() for s in temp_var_names.split(",")]

    from loopy.kernel.data import AddressSpace
    if isinstance(address_space, str):
        try:
            address_space = getattr(AddressSpace, address_space.upper())
        except AttributeError as err:
            raise LoopyError("address_space '%s' unknown" % address_space) from err

    if not isinstance(address_space, int) or address_space not in [
            AddressSpace.PRIVATE,
            AddressSpace.LOCAL,
            AddressSpace.GLOBAL]:
        raise LoopyError("invalid address_space '%s'" % address_space)

    new_temp_vars = kernel.temporary_variables.copy()
    for tv_name in temp_var_names:
        try:
            tv = new_temp_vars[tv_name]
        except KeyError:
            raise LoopyError("temporary '%s' not found" % tv_name) from None

        new_temp_vars[tv_name] = tv.copy(address_space=address_space)

    return kernel.copy(temporary_variables=new_temp_vars)


def set_temporary_scope(kernel, temp_var_names, address_space):
    warn("set_temporary_scope is deprecated and will stop working in "
            "July 2022. Use set_temporary_address_space instead.",
            DeprecationWarning, stacklevel=2)

    return set_temporary_address_space(kernel, temp_var_names, address_space)

# }}}


# {{{ reduction_arg_to_subst_rule

@for_each_kernel
def reduction_arg_to_subst_rule(
        kernel, inames, insn_match=None, subst_rule_name=None):
    if isinstance(inames, str):
        inames = [s.strip() for s in inames.split(",")]

    inames_set = frozenset(inames)

    substs = kernel.substitutions.copy()

    var_name_gen = kernel.get_var_name_generator()

    def map_reduction(expr, rec, nresults=1):
        if frozenset(expr.inames) != inames_set:
            return type(expr)(
                    operation=expr.operation,
                    inames=expr.inames,
                    expr=rec(expr.expr),
                    allow_simultaneous=expr.allow_simultaneous)

        if subst_rule_name is None:
            subst_rule_prefix = "red_%s_arg" % "_".join(inames)
            my_subst_rule_name = var_name_gen(subst_rule_prefix)
        else:
            my_subst_rule_name = subst_rule_name

        if my_subst_rule_name in substs:
            raise LoopyError("substitution rule '%s' already exists"
                    % my_subst_rule_name)

        from loopy.kernel.data import SubstitutionRule
        substs[my_subst_rule_name] = SubstitutionRule(
                name=my_subst_rule_name,
                arguments=tuple(inames),
                expression=expr.expr)

        from pymbolic import var
        iname_vars = [var(iname) for iname in inames]

        return type(expr)(
                operation=expr.operation,
                inames=expr.inames,
                expr=var(my_subst_rule_name)(*iname_vars),
                allow_simultaneous=expr.allow_simultaneous)

    from loopy.symbolic import ReductionCallbackMapper
    cb_mapper = ReductionCallbackMapper(map_reduction)

    from loopy.kernel.data import MultiAssignmentBase

    new_insns = []
    for insn in kernel.instructions:
        if not isinstance(insn, MultiAssignmentBase):
            new_insns.append(insn)
        else:
            new_insns.append(insn.copy(expression=cb_mapper(insn.expression)))

    return kernel.copy(
            instructions=new_insns,
            substitutions=substs)

# }}}


# {{{ add_padding_to_avoid_bank_conflicts

# experimental not exported/documented for now
@for_each_kernel
def add_padding_to_avoid_bank_conflicts(kernel, device):
    import pyopencl as cl
    import pyopencl.characterize as cl_char

    new_temp_vars = {}

    from loopy.kernel.data import AddressSpace

    lmem_size = cl_char.usable_local_mem_size(device)
    for temp_var in kernel.temporary_variables.values():
        if temp_var.address_space != AddressSpace.LOCAL:
            new_temp_vars[temp_var.name] = \
                    temp_var.copy(storage_shape=temp_var.shape)
            continue

        if not temp_var.shape:
            # scalar, no need to mess with storage shape
            new_temp_vars[temp_var.name] = temp_var
            continue

        other_loctemp_nbytes = [
                tv.nbytes
                for tv in kernel.temporary_variables.values()
                if tv.address_space == AddressSpace.LOCAL
                and tv.name != temp_var.name]

        storage_shape = temp_var.storage_shape

        if storage_shape is None:
            storage_shape = temp_var.shape

        storage_shape = list(storage_shape)

        # sizes of all dims except the last one, which we may change
        # below to avoid bank conflicts
        from pytools import product

        if device.local_mem_type == cl.device_local_mem_type.GLOBAL:
            # FIXME: could try to avoid cache associativity disasters
            new_storage_shape = storage_shape

        elif device.local_mem_type == cl.device_local_mem_type.LOCAL:
            min_mult = cl_char.local_memory_bank_count(device)
            good_incr = None
            new_storage_shape = storage_shape
            min_why_not = None

            for increment in range(storage_shape[-1]//2):

                test_storage_shape = storage_shape[:]
                test_storage_shape[-1] = test_storage_shape[-1] + increment
                new_mult, why_not = cl_char.why_not_local_access_conflict_free(
                        device, temp_var.dtype.itemsize,
                        temp_var.shape, test_storage_shape)

                # will choose smallest increment 'automatically'
                if new_mult < min_mult:
                    new_lmem_use = (sum(other_loctemp_nbytes)
                            + temp_var.dtype.itemsize*product(test_storage_shape))
                    if new_lmem_use < lmem_size:
                        new_storage_shape = test_storage_shape
                        min_mult = new_mult
                        min_why_not = why_not
                        good_incr = increment

            if min_mult != 1:
                from loopy.diagnostic import LoopyAdvisory
                warn("could not find a conflict-free mem layout "
                        "for local variable '%s' "
                        "(currently: %dx conflict, increment: %s, reason: %s)"
                        % (temp_var.name, min_mult, good_incr, min_why_not),
                        LoopyAdvisory, stacklevel=4)
        else:
            warn("unknown type of local memory", stacklevel=4)

            new_storage_shape = storage_shape

        new_temp_vars[temp_var.name] = temp_var.copy(
                storage_shape=tuple(new_storage_shape))

    return kernel.copy(temporary_variables=new_temp_vars)

# }}}


# {{{ allocate_temporaries_for_base_storage

@dataclass(frozen=True)
class _BaseStorageInfo:
    name: str
    next_offset: ArithmeticExpression
    approx_nbytes: int | None = None


def _sym_max(a: Expression, b: Expression) -> Expression:
    from numbers import Number
    if isinstance(a, Number) and isinstance(b, Number):
        return max(a, b)
    else:
        from pymbolic.primitives import Max
        return Max((a, b))


@for_each_kernel
def allocate_temporaries_for_base_storage(kernel: LoopKernel,
        only_address_space: int | None = None,
        aliased=True,
        max_nbytes: int | None = None,
        ) -> LoopKernel:
    from pytools import product

    new_tvs = dict(kernel.temporary_variables)
    made_changes = False

    vng = kernel.get_var_name_generator()

    name_aspace_dtype_to_bsi: dict[
            tuple[str, AddressSpace, LoopyType], _BaseStorageInfo] = {}

    for tv in sorted(
            kernel.temporary_variables.values(),
            key=lambda key_tv: key_tv.name):
        if tv.base_storage and tv.initializer:
            raise LoopyError(
                    f"Temporary '{tv.name}' has both base_storage "
                    "and an initializer. That's not allowed.")
        if tv.offset and not tv.base_storage:
            raise LoopyError(
                    f"Temporary '{tv.name}' has an offset and no base_storage. "
                    "That's not allowed.")
        if not isinstance(tv.dtype, LoopyType):
            raise LoopyError(
                f"Dtype of temporary '{tv.name}' "
                " is not inferred. Call lp.infer_unknown_types"
                " first.")

        if (tv.base_storage
                and tv.base_storage not in kernel.temporary_variables
                and (
                    only_address_space is None
                    or tv.address_space == only_address_space)):
            made_changes = True
            assert isinstance(tv.dtype, LoopyType)

            if tv.address_space is auto:
                raise LoopyError("When allocating base storage for temporary "
                        f"'{tv.name}', the address space of the temporary "
                        "was not yet determined (set to 'auto').")

            assert isinstance(tv.shape, tuple)
            ary_size = product(si for si in tv.shape)
            if isinstance(ary_size, (int, np.integer)):
                approx_array_nbytes = tv.dtype.numpy_dtype.itemsize * ary_size
            else:
                # FIXME: Could use approximate values of ValueArgs
                approx_array_nbytes = 0

            bs_key = (tv.base_storage,
                      cast("AddressSpace", tv.address_space), tv.dtype)
            bsi = name_aspace_dtype_to_bsi.get(bs_key)

            if bsi is None or (
                    # are we out of space?
                    not aliased
                    and max_nbytes is not None
                    and bsi.approx_nbytes is not None
                    and bsi.approx_nbytes + approx_array_nbytes > max_nbytes):
                bsi = name_aspace_dtype_to_bsi[bs_key] = _BaseStorageInfo(
                        name=vng(tv.base_storage),
                        next_offset=0,
                        approx_nbytes=None if aliased else 0)

                new_tvs[bsi.name] = TemporaryVariable(
                    name=bsi.name,
                    dtype=tv.dtype,
                    shape=(0,),
                    address_space=tv.address_space)

            new_tvs[tv.name] = tv.copy(
                base_storage=bsi.name,
                offset=bsi.next_offset,
                _base_storage_access_may_be_aliasing=(
                    aliased if tv._base_storage_access_may_be_aliasing is None
                    else tv._base_storage_access_may_be_aliasing))

            bs_tv = new_tvs[bsi.name]
            bs_size: ArithmeticExpression
            bs_size, = assert_tuple(bs_tv.shape)
            if aliased:
                new_bs_size = _sym_max(bs_size, ary_size)
            else:
                new_bs_size = bs_size + ary_size

                assert bsi.approx_nbytes is not None
                name_aspace_dtype_to_bsi[bs_key] = replace(bsi,
                    next_offset=bsi.next_offset + ary_size,
                    approx_nbytes=bsi.approx_nbytes + approx_array_nbytes)

            new_tvs[bsi.name] = new_tvs[bsi.name].copy(shape=(new_bs_size,))

    if made_changes:
        return kernel.copy(temporary_variables=new_tvs)
    else:
        return kernel

# }}}

# vim: foldmethod=marker



__copyright__ = "Copyright (C) 2016 Matt Wala"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging
from pickle import dumps, loads

import pytest

import loopy as lp
from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2  # noqa: F401


logger = logging.getLogger(__name__)


def test_kernel_pickling_and_hashing():
    knl = lp.make_kernel("{[i]: 0<=i<10}",
                         """
                         y[i] = i
                         """)
    from loopy.tools import LoopyKeyBuilder
    reconst_knl = loads(dumps(knl))
    assert LoopyKeyBuilder()(knl) == LoopyKeyBuilder()(reconst_knl)


def test_set_trie():
    from loopy.kernel.tools import SetTrie

    s = SetTrie()
    s.add_or_update({1, 2, 3})
    s.add_or_update({4, 2, 1})
    s.add_or_update({1, 5})

    result = []
    s.descend(lambda prefix: result.extend(prefix))
    assert result == [1, 2, 3, 4, 5]

    with pytest.raises(ValueError):
        s.add_or_update({1, 4})


class PickleDetector:
    """Contains a class attribute which flags if any instance was unpickled.
    """

    @classmethod
    def reset(cls):
        cls.instance_unpickled = False

    def __getstate__(self):
        return {"state": self.state}

    def __setstate__(self, state):
        self.__class__.instance_unpickled = True
        self.state = state["state"]


class PickleDetectorForLazilyUnpicklingDict(PickleDetector):
    instance_unpickled = False

    def __init__(self):
        self.state = None


def test_lazily_unpickling_dict():
    from loopy.tools import LazilyUnpicklingDict

    cls = PickleDetectorForLazilyUnpicklingDict
    mapping = LazilyUnpicklingDict({0: cls()})

    assert not cls.instance_unpickled

    pickled_mapping = dumps(mapping)

    # {{{ test lazy loading

    mapping = loads(pickled_mapping)
    assert not cls.instance_unpickled
    list(mapping.keys())
    assert not cls.instance_unpickled
    assert isinstance(mapping[0], cls)
    assert cls.instance_unpickled

    # }}}

    # {{{ conversion

    cls.reset()
    mapping = loads(pickled_mapping)
    dict(mapping)
    assert cls.instance_unpickled

    # }}}

    # {{{ test multi round trip

    mapping = loads(dumps(loads(pickled_mapping)))
    assert isinstance(mapping[0], cls)

    # }}}

    # {{{ test empty map

    mapping = LazilyUnpicklingDict({})
    mapping = loads(dumps(mapping))
    assert len(mapping) == 0

    # }}}


class PickleDetectorForLazilyUnpicklingList(PickleDetector):
    instance_unpickled = False

    def __init__(self):
        self.state = None


def test_lazily_unpickling_list():
    from loopy.tools import LazilyUnpicklingList

    cls = PickleDetectorForLazilyUnpicklingList
    lst = LazilyUnpicklingList([cls()])
    assert not cls.instance_unpickled

    pickled_lst = dumps(lst)

    # {{{ test lazy loading

    lst = loads(pickled_lst)
    assert not cls.instance_unpickled
    assert isinstance(lst[0], cls)
    assert cls.instance_unpickled

    # }}}

    # {{{ conversion

    cls.reset()
    lst = loads(pickled_lst)
    list(lst)
    assert cls.instance_unpickled

    # }}}

    # {{{ test multi round trip

    lst = loads(dumps(loads(dumps(lst))))
    assert isinstance(lst[0], cls)

    # }}}

    # {{{ test empty list

    lst = LazilyUnpicklingList([])
    lst = loads(dumps(lst))
    assert len(lst) == 0

    # }}}


class PickleDetectorForLazilyUnpicklingListWithEqAndPersistentHashing(
        PickleDetector):
    instance_unpickled = False

    def __init__(self, comparison_key):
        self.state = comparison_key

    def __repr__(self):
        return repr(self.state)

    def update_persistent_hash(self, key_hash, key_builder):
        key_builder.rec(key_hash, repr(self))


def test_lazily_unpickling_list_eq_and_persistent_hashing():
    from loopy.tools import LazilyUnpicklingListWithEqAndPersistentHashing

    cls = PickleDetectorForLazilyUnpicklingListWithEqAndPersistentHashing

    # {{{ test comparison of a pair of lazy lists

    lst0 = LazilyUnpicklingListWithEqAndPersistentHashing(
            [cls(0), cls(1)],
            eq_key_getter=repr,
            persistent_hash_key_getter=repr)
    lst1 = LazilyUnpicklingListWithEqAndPersistentHashing(
            [cls(0), cls(1)],
            eq_key_getter=repr,
            persistent_hash_key_getter=repr)

    assert not cls.instance_unpickled

    assert lst0 == lst1
    assert not cls.instance_unpickled

    lst0 = loads(dumps(lst0))
    lst1 = loads(dumps(lst1))

    assert lst0 == lst1
    assert not cls.instance_unpickled

    lst0.append(cls(3))
    lst1.append(cls(2))

    assert lst0 != lst1

    # }}}

    # {{{ comparison with plain lists

    lst = [cls(0), cls(1), cls(3)]

    assert lst == lst0
    assert lst0 == lst
    assert not cls.instance_unpickled

    # }}}

    # {{{ persistent hashing

    from loopy.tools import LoopyKeyBuilder
    kb = LoopyKeyBuilder()

    assert kb(lst0) == kb(lst)
    assert not cls.instance_unpickled

    # }}}


def test_optional():
    from loopy import Optional

    # {{{ test API

    opt = Optional()
    assert not opt.has_value
    with pytest.raises(AttributeError):
        opt.value  # noqa: B018

    opt = Optional(1)
    assert opt.has_value
    assert 1 == opt.value

    assert Optional(1) == Optional(1)
    assert Optional(1) != Optional(2)
    assert Optional() == Optional()
    assert Optional() != Optional(1)

    # }}}

    # {{{ test pickling

    import pickle

    assert not pickle.loads(pickle.dumps(Optional())).has_value
    assert pickle.loads(pickle.dumps(Optional(1))).value == 1

    # }}}

    # {{{ test key builder

    from loopy.tools import LoopyKeyBuilder
    kb = LoopyKeyBuilder()
    kb(Optional())
    kb(Optional(None))

    # }}}


@lp.memoize_on_disk
def very_costly_transform(knl, iname):
    from time import sleep
    sleep(5)
    return lp.split_iname(knl, iname, 4)


def test_memoize_on_disk():
    if not lp.CACHING_ENABLED:
        # if caching is disabled => don't test the caching behavior
        pytest.skip("cannot test memoization if caching disabled")

    knl = lp.make_kernel("{[i]: 0<=i<10}",
                         """
                         y[i] = i
                         """)

    from time import time
    uncached_knl = very_costly_transform(knl, "i")

    start = time()
    cached_knl = very_costly_transform(knl, "i")
    end = time()
    assert (end - start) < 4
    assert cached_knl == uncached_knl


@lp.memoize_on_disk
def get_twice_of_pym_expr(expr):
    from time import sleep
    sleep(2)
    return 2 * expr


def test_memoize_on_disk_with_pym_expr():
    if not lp.CACHING_ENABLED:
        # if caching is disabled => don't test the caching behavior
        pytest.skip("cannot test memoization if caching disabled")

    from pymbolic import parse
    expr = parse("a[i] + b[i]")

    from time import time
    uncached_result = get_twice_of_pym_expr(expr)

    start = time()
    cached_result = get_twice_of_pym_expr(expr)
    end = time()
    assert (end - start) < 1.5
    assert cached_result == uncached_result


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])

# vim: foldmethod=marker
